// SI4844 Arduino Radio
// Программа к статье:
// Собираем управляемый Arduino радиоприемник с AM на СВ/КВ и FM в УКВ диапазоне
// http://www.radioprog.ru/post/232


#include <Wire.h>           // для работы Arduino с шиной I2C
// предполагается, что библиотеки клавиатуры и дисплея уже установлены
#include <Keypad.h>         // для мембранной клавиатуры
#include <LCD5110_Basic.h>  // для дисплея Nokia

// определения
// I2C адрес микросхемы зашит жестко
#define Si_ADDRESS 0x11
// команды Si из документации
#define ATDD_POWER_DOWN 0x11
#define ATDD_POWER_UP   0xE1
#define ATDD_XOWAIT     0x40
#define SET_PROPERTY    0x12
#define RX_VOLUME       0x40
#define ATDD_GET_STATUS 0xE0

#define TONEH           0x40
#define TONEL           0x02
#define MUTEH           0x40
#define MUTEL           0x01

// глобальные переменные
// дисплей
LCD5110 myGLCD(7,6,5,3,4);
extern uint8_t SmallFont[];
extern uint8_t MediumNumbers[];
extern uint8_t BigNumbers[];
// клавиатура
const byte ROWS = 4; // четыре строки
const byte COLS = 3; // три столбца
char keys[ROWS][COLS] = {
  {'1','2','3'},
  {'4','5','6'},
  {'7','8','9'},
  {'*','0','#'}
};
// выводы подключения клавиатуры к Arduino
byte rowPins[ROWS] = {8,9,10,11}; // подключить к выводам строк клавиатуры
byte colPins[COLS] = {13,14,15};  // подключить к выводам столбцов клавиатуры
// получить объект
Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );
char key;           // хранит нажатую клавишу
// выводы Arduino D2=прерывание, D12=сброс
const byte INTpin=2;
const byte RESETpin=12;
// для прерывания
volatile byte IRQflag;
// Диапазоны по умолчанию в FM/AM/SW - диапазоны, с которых мы начинаем при переключении режима
byte d_FM_band=8;
byte d_AM_band=22;
byte d_SW_band=31;
byte status_rsp[4];     // хранит байты отклика на команду
byte current_band=0;    // 0=FM 1-AM 2=SW  при включении питания мы начинаем с 0
byte current_bandix;    // индекс текущего диапазона
byte current_mute=0;    // 0=выкл 1=вкл (выключен звук)
byte current_font;      // 0=маленький 1=средний
byte current_vol=50;    // значение при включении
byte current_BT=4;      // значение при включени 4 или 3 FM/AM
byte power_mode=1;      // включены ли мы (1=включен 0=выключен)
byte mute_mode=0;       // показывает режим выключения звука (1=вкл 0=выкл)
char buf2[3];           // для itoa преобразования до двух символов
int d1,d2,d3,d4;        // для простоты печати
// Строки для простоты и быстроты печати частотных диапазонов
char * BandRange[]={"  87-108   ","  87-108   ","  87-108   ","  87-108   ","  86.5-109 ", "  86.5-109 ", "  86.5-109 ","  86.5-109 ",
"87.3-108.25","87.3-108.25","87.3-108.25","87.3-108.25","   76-90   ","   76-90   ","   76-90   ","   76-90   ","   64-87   ","   64-87   ",
"   64-87   ","   64-87   ","  520-1710 ","  522-1620 ","  504-1665 ","  520-1730 ","  510-1750 ","  5.6-6.4  ","  5.95-6.2 ","  6.8-7.6  ",
"  7.1-7.6  ","   9.2-10  ","  9.2-9.9  ","11.45-12.25"," 11.6-12.2 "," 13.4-14.2 ","13.57-13.87","  15-15.9  "," 15.1-15.8 ","  17.1-18  ",
" 17.48-17.9","  21.2-22  ","21.45-21.85"};

void setup() 
{
  // настройка выводов и прерываний
  // прерывание активно HIGH
  pinMode (INTpin, INPUT);
  // вывод сброса
  pinMode(RESETpin,OUTPUT);
  digitalWrite(RESETpin,HIGH);
  // ISR
  attachInterrupt(0,SiISR,RISING);
  // настройка дисплея
  myGLCD.InitLCD();
  display_template();
  // переключиться в FM диапазон с индексом диапазона по умолчанию
  SiReset();
  current_bandix=d_FM_band;
  SiSetBand(); 
  get_status();
  // включение завершено, поэтому получим значения от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency();  
  go_current();   // установит по умолчанию настройки громкости, тона, выключения звука
}

void loop() 
{
  // Проверка изменений настроек и нажатий клавиш

  if(IRQflag)
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(ATDD_GET_STATUS);  
    Wire.endTransmission();  
    Wire.requestFrom(Si_ADDRESS, 0x04);
    status_rsp[0]=Wire.read();
    status_rsp[1]=Wire.read();
    status_rsp[2]=Wire.read();
    status_rsp[3]=Wire.read();
    display_frequency();  
    IRQflag=0;
  }
  // получить и обработать нажатие клавиши, если оно было
  if(key=keypad.getKey())
  {
    switch(key)
    {
      case '1':
        Go_AM();
        break;
      case '2':
        Go_FM();
        break;
      case '3':
        Go_SW();
        break;
      case '4':
        Vol_up();
        break;
      case '7':
        Vol_down();
        break;  
      case '5':
        Band_up();
        break;  
      case '8':
        Band_down();
        break;  
      case '6':
        BT_up();
        break;   
      case '9':
        BT_down();
        break;
      case '0':
        Mute();
        break; 
      default:
        break;  
    }
  }
}

void SiReset()
{
  // сбросить Si4844 и вернуться 10 кГц на I2C
  I2C_10kHz();
  IRQflag=0;
  digitalWrite(RESETpin,LOW);
  delayMicroseconds(200);
  digitalWrite(RESETpin,HIGH);
  delayMicroseconds(200);
  // ждать нарастающего фронта прерывания
  while(!IRQflag);
  // ждем 2.5 мс
  delayMicroseconds(2500);
}

void SiGetStatus()
{
  // отправить команду состояния и ждать 2 мс перед проверкой байта отклика
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(ATDD_GET_STATUS);  
  Wire.endTransmission();  
  delayMicroseconds(2000);
  // проверка байта отклика
  Wire.requestFrom(Si_ADDRESS, 0x04);
  for(int x=0;x<4;x++)
  {
    status_rsp[x]=Wire.read();
  }
}

void SiSetBand()
{
  // предполагаем, что current_band установлен
  // включаем и получаем состояние о настройках status_rsp[0-4]
  byte cb;
  cb=current_bandix;

  bitSet(cb,7);                // бит 7=1 для внешнего кварца
  bitClear(cb,6);              // бит 6=0 для обычного времени кварца
                               // arg[2]-[7] не используются в этом режиме
  IRQflag=0;
  // послать
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(ATDD_POWER_UP);  
  Wire.write(cb);
  Wire.endTransmission();  
  // ждать прерывания, говорящего о готовности настройки частоты
  delayMicroseconds(2500);
  while(!IRQflag);
  // ждать 2,5 мс, т.к. у нас прерывание по нарастающему фронту
  delayMicroseconds(2500);
  I2C_50kHz();
  // послать получение статуса и ждать 2 мс перед проверкой байта отклика
  get_status();
}

void display_template()
{
  myGLCD.clrScr();
  myGLCD.setFont(SmallFont);
  myGLCD.print("BAND",32,0);
  myGLCD.print("Vol:",0,16);
  myGLCD.print("B/T:",48,16);
  myGLCD.print("mHz",66,32);
  myGLCD.setFont(MediumNumbers);
  current_font=1;
  myGLCD.print("123.4",0,24);
}

void display_band_index()
{
  // предполагаем, current_bandix точен
  // показать значение индекса диапазона
  if (current_font==1)
  {
    // установить маленький шрифт
    myGLCD.setFont(SmallFont);
    current_font=0;
  }
  myGLCD.print(BandRange[current_bandix],8,8);
  itoa(current_bandix,buf2,10);
  if(current_bandix<10)
  { 
    // с отступом
    myGLCD.print(" ",64,0);
    myGLCD.print(buf2,70,0);
  }
  else
  {
    myGLCD.print(buf2,64,0);
  }
}

void display_band()
{
  // предполагаем, current_bandix точен 0=FM 1-AM 2=SW
  // также изменить k/M в частоте
  if (current_font==1)
  {
    // set small font
    myGLCD.setFont(SmallFont);
    current_font=0;
  }
  switch(current_band)
  {
    case 0:
      myGLCD.print("FM",8,0);
      myGLCD.print("M",66,32);
      break;
    case 1:
      myGLCD.print("AM",8,0);
      myGLCD.print("k",66,32);
      break;
    case 2:
      myGLCD.print("SW",8,0);
      myGLCD.print("M",66,32);
      break;
    default:
      // ошибка
      myGLCD.print("XX",8,0);
      break;  
  }  
}

void display_frequency()
{
  int add;
  // предполагаем что current_band и status_rsp[2] и status_rsp[3] соответствуют
  // рассчитать 4 цифры и получить их строковые эквиваленты для дисплея
  // Если CHFREQ[15] установлен в диапазоне SW, он показывает  добавление 5 кГц к частоте,
  // т.е., частота=частота + 0.005 кГц
  // Сейчас это не реализовано, но это здесь, если я захочу сделать что-нибудь с этим позже 
  add=0;
  if((current_band==2) && (bitRead(status_rsp[2],7)==1))
  {
    add=1;
    bitClear(status_rsp[2],7);
  }
  d1=status_rsp[2]>>4;
  d2=status_rsp[2]%16;
  d3=status_rsp[3]>>4;
  d4=status_rsp[3]%16;
  if (current_font==0)
  {
    // установить средний шрифт
    myGLCD.setFont(MediumNumbers);
    current_font=1;
  }
  switch(current_band)
  {
    // FM это МГц 012.3
    case 0:
      myGLCD.printNumI(d1,0,24);
      myGLCD.printNumI(d2,12,24);
      myGLCD.printNumI(d3,24,24);
      myGLCD.print(".",36,24);
      myGLCD.printNumI(d4,48,24);
      // проверить маркер стерео, если мы в FM
      // необходимо сделать это более эффективно
      myGLCD.setFont(SmallFont);
      current_font=0;
      if(bitRead(status_rsp[0],2))
      {
        myGLCD.print("stereo",0,40);
      }
      else
      {
        myGLCD.print("      ",0,40);
      }
      myGLCD.setFont(MediumNumbers);
      current_font=1;
      break;
    case 1:
      // AM это 1234 кГц с отступом
      myGLCD.print(" ",0,24);
      myGLCD.printNumI(d1,12,24);
      myGLCD.printNumI(d2,24,24);
      myGLCD.printNumI(d3,36,24);
      myGLCD.printNumI(d4,48,24);
      break;    
    case 2:  
      // SW это МГц 01.23     
      myGLCD.printNumI(d1,0,24);
      myGLCD.printNumI(d2,12,24);
      myGLCD.print(".",24,24);
      myGLCD.printNumI(d3,36,24);
      myGLCD.printNumI(d4,48,24);
      break;
    default:
      // ошибка
      myGLCD.print("X",0,24);
      myGLCD.print("X",12,24);
      myGLCD.print("X",24,24);
      myGLCD.print("X",36,24);
      myGLCD.print("X",48,24);
      break;
  }
}

void display_mute()
{
  // пишет или стирает 'mute' на дисплее в зависимости
  // от состояния mute_mode
  if (current_font==1)
  {
    // установить маленький шрифт
    myGLCD.setFont(SmallFont);
    current_font=0;
  }
  if(mute_mode==1)
  {
    // написать mute
    myGLCD.print("mute",60,40);
  }
  else
  {
    myGLCD.print("    ",60,40); 
  }
}

void display_vol()
{
  // предполагаем, что current_vol соответствует
  if (current_font==1)
  {
    // установить маленький шрифт
    myGLCD.setFont(SmallFont);
    current_font=0;
  }
  itoa(current_vol,buf2,10);
  if(current_vol<10)
  { // отступ
    myGLCD.print(" ",24,16);
    myGLCD.print(buf2,30,16);
  }
  else
  {
    myGLCD.print(buf2,24,16);
  }
}

void display_BT()
{
  // предполагаем, что current_BT соответствует
  if (current_font==1)
  {
    // установить маленький шрифт
    myGLCD.setFont(SmallFont);
    current_font=0;
  }
  myGLCD.printNumI(current_BT,72,16);
}

void Mute()
{
  // если mute_mode<>0 включить звук и установить mute_mode в 0
  // если mute_mode=0 выключить звук и установить mute_mode в 1
  if(mute_mode==0)
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(MUTEH);              // ARG2 0x40
    Wire.write(MUTEL);              // ARG3 0x01
    Wire.write(0x00);               // ARG4    
    Wire.write(0x03);               // ARG5 выключить L и R
    Wire.endTransmission(); 
    mute_mode=1;
    display_mute();
  }
  else
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(MUTEH);              // ARG2 0x40
    Wire.write(MUTEL);              // ARG3 0x01
    Wire.write(0x00);               // ARG4 
    Wire.write(0x00);               // ARG5 включить L и R
    Wire.endTransmission(); 
    mute_mode=0;
    display_mute();
  }
}

void BT_up()
{
  // мы работаем в стандартном Mixed Mode 2
  // предполагаем, что current_BT соответствует
  if (current_BT <8)
  {
    current_BT++;
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(TONEH);              // ARG2 0x40
    Wire.write(TONEL);              // ARG3 0x02
    Wire.write(0x00);               // ARG4
    Wire.write(current_BT);         // ARG5 это бас/тембр 0-7*
    Wire.endTransmission(); 
    display_BT();
  }
}

void BT_down()
{
  // мы работаем в стандартном Mixed Mode 2
  // предполагаем, что current_BT соответствует
  if (current_BT >0)
  {
    current_BT--;
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(TONEH);              // ARG2 0x40
    Wire.write(TONEL);              // ARG3 0x02
    Wire.write(0x00);               // ARG4
    Wire.write(current_BT);         // ARG5 это бас/тембр 0-7*
    Wire.endTransmission(); 
    display_BT();
  }
}

void Vol_up()
{
  // мы работаем в стандартном Mixed Mode 2
  // предполагаем, что current_vol соответствует
  if (current_vol <63)
  {
    current_vol++;
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);         // ARG1
    Wire.write(RX_VOLUME);    // ARG2 RX_VOLUME 0x40
    Wire.write(0x00);         // ARG3
    Wire.write(0x00);         // ARG4
    Wire.write(current_vol);  // ARG5 реальная громкость 0-63
    Wire.endTransmission(); 
    display_vol();
  }
}

void Vol_down()
{
  // мы работаем в стандартном Mixed Mode 2
  // предполагаем, что current_vol соответствует
  if (current_vol >0)
  {
    current_vol--;
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);         // ARG1
    Wire.write(RX_VOLUME);    // ARG2 RX_VOLUME 0x40
    Wire.write(0x00);         // ARG3
    Wire.write(0x00);         // ARG4
    Wire.write(current_vol);  // ARG5 реальная громкость 0-63
    Wire.endTransmission(); 
    display_vol();
  }
}

void Go_FM()
{
  // перелючиться в диапазон FM со стандартным band_idx
  SiReset();
  current_bandix=d_FM_band;
  SiSetBand(); 
  get_status();
  // включение завершено, поэтому получаем данные от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency(); 
  go_current();  
}

void Go_AM()
{
  // перелючиться в диапазон AM со стандартным band_idx
  myGLCD.setFont(SmallFont);
  current_font=0;
  myGLCD.print("      ",0,40);  // очистить маркер стерео, если он был
  SiReset();
  current_bandix=d_AM_band;
  SiSetBand(); 
  // в тестировании это никогда не повторяется
  get_status();
  // включение завершено, поэтому получаем данные от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency();   
  go_current();
}

void Go_SW()
{
  // перелючиться в диапазон SW со стандартным band_idx
  myGLCD.setFont(SmallFont);
  current_font=0;
  myGLCD.print("      ",0,40);  // очистить маркер стерео, если он был
  SiReset();
  current_bandix=d_SW_band;
  SiSetBand(); 
  get_status();
  // включение завершено, поэтому получаем данные от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency();   
  go_current();
}

void go_current()
{
  // восстановить текущие настройки, когда мы переключаем AM/FM/SW
  // Установить громкость в значение current_vol
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(SET_PROPERTY);  
  Wire.write(0x00);         // ARG1
  Wire.write(RX_VOLUME);    // ARG2 RX_VOLUME 0x40
  Wire.write(0x00);         // ARG3
  Wire.write(0x00);         // ARG4
  Wire.write(current_vol);  // ARG5 реальная громкость 0-63
  Wire.endTransmission(); 
  display_vol();
  // Установить B/T в значение current_BT
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(SET_PROPERTY);  
  Wire.write(0x00);               // ARG1
  Wire.write(TONEH);              // ARG2 0x40
  Wire.write(TONEL);              // ARG3 0x02
  Wire.write(0x00);               // ARG4
  Wire.write(current_BT);         // ARG5 это бас/тембр 0-7*
  Wire.endTransmission(); 
  display_BT();
  // включить или выключить звук
  if(mute_mode==1)
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(MUTEH);              // ARG2 0x40
    Wire.write(MUTEL);              // ARG3 0x01
    Wire.write(0x00);               // ARG4    
    Wire.write(0x03);               // ARG5 выключить L и R
    Wire.endTransmission(); 
  }
  else
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(SET_PROPERTY);  
    Wire.write(0x00);               // ARG1
    Wire.write(MUTEH);              // ARG2 0x40
    Wire.write(MUTEL);              // ARG3 0x01
    Wire.write(0x00);               // ARG4 
    Wire.write(0x00);               // ARG5 включить L и R
    Wire.endTransmission(); 
  }
  display_mute();
}


void Band_up()
{
  // увеличить индекс диапазона на 1
  // предполагаем, что current_band и current_bandix соответствуют
  switch(current_band)
  {
    case 0: // FM это 0 19
      if(current_bandix<19)
      {
        current_bandix++;     
      }
      break;
    case 1:  // AM это 20-24
      if(current_bandix<24)
      {
        current_bandix++;     
      }
      break;
    case 2:  //  SW это 25-40
      if(current_bandix<40)
      {
        current_bandix++;     
      }  
      break;      
    default:
      // ошибка - ничего не делать
      break;   
  }
  SiReset();
  SiSetBand(); 
  get_status();
  // включение завершено, поэтому получаем данные от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency();  
  go_current();
}

void Band_down()
{
  // уменьшить индекс диапазона на 1
  // предполагаем, что current_band и current_bandix соответствуют
  switch(current_band)
  {
    case 0: // FM это 0 19
      if(current_bandix>0)
      {
        current_bandix--;     
      }
      break;
    case 1:  // AM это 20-24
      if(current_bandix>20)
      {
        current_bandix--;     
      }
      break;
    case 2:  //  SW это 25-40
      if(current_bandix>25)
      {
        current_bandix--;     
      }  
      break;      
    default:
      // ошибка - ничего не делать
      break;   
  }
  SiReset();
  SiSetBand(); 
  get_status();
  // включение завершено, поэтому получаем данные от чипа
  current_band=(status_rsp[1]>>6);
  display_band();
  current_bandix=(status_rsp[1]&B00111111);
  display_band_index();
  display_frequency();
  go_current();
}

void get_status()
{
  // выполнять SiGetStatus, пока не получим корректную частоту в status_rsp
  // примечание: в тестировании мы не использовали цикл, но это хорошая практика
  // также, сюда может быть добавлена процедура выхода после определенного количества попыток
  Waitforgo:               
    SiGetStatus(); 
    if( ((bitRead(status_rsp[0],4)==0) || ( (status_rsp[2]==0) && (status_rsp[3]==0))) )
    {
      goto Waitforgo;       
    }
}

// Примечание: Это функции установки скорости I2C для 8МГц , 3.3 В Pro Mini
// Для 16 МГц UNO или других плат Arduino их необходимо изменить
void I2C_10kHz()
{
  // установить скорость I2C на 10 кГц на чипе 8 МГц
  TWBR=98;
  TWSR|=bit(TWPS0);
}

void I2C_50kHz()
{
  // установить скорость I2C на 50 кГц на чипе 8 МГц
  TWBR=18;
  TWSR|=bit(TWPS0);
} 

// Функция обработки прерывания
void SiISR()
{
    IRQflag=1;
}






