// SI4844 Quick Test
// Простая тестовая программа к статье
// Собираем управляемый Arduino радиоприемник с AM на СВ/КВ и FM в УКВ диапазоне
// http://www.radioprog.ru/post/232

#include <Wire.h>
// I2C адрес микросхемы зашит жестко
#define Si_ADDRESS 0x11
// команды Si
#define ATDD_POWER_DOWN 0x11
#define ATDD_POWER_UP   0xE1
#define GET_REV         0x10
#define ATDD_XOSCEN     0x80
#define ATDD_XOWAIT     0x40
#define ATDD_GET_STATUS 0xE0

// глобальные переменные
// выводы Arduino D2=прерывание, D12=сброс
const byte INTpin=2;
const byte RESETpin=12;
byte status_rsp[9];
volatile byte IRQflag;

void setup() 
{
  // настройка выводов и прерываний
  Serial.begin(115200); 
  // прерывание активно HIGH
  pinMode (INTpin, INPUT);
  // сброс
  pinMode(RESETpin,OUTPUT);
  digitalWrite(RESETpin,HIGH);
  attachInterrupt(0,SiISR,RISING);
}

void loop() 
{
  byte band;
  byte bandmode;
  // Примечание: мы использует режим "Определение диапазона контроллером системы" 
  delay(2000);  // ждать открытия монитора последовательного порта
  Serial.println("*** Si4844 Quick Test **");
  Serial.println("...setting I2C speed to 10kHz");
  I2C_10kHz();
  Serial.println("...resetting Si4844");
  IRQflag=0;
  SiReset();
  // ждать нарастающего фронта прерывания
  while(!IRQflag);
  // ждать 2.5 мс
  delayMicroseconds(2500);
  Serial.println("Si4844 reset complete");
  band=0;             // FM диапазон 0 - 87-108 МГц
  bitSet(band,7);     // бит 7=1 для внешнего кварца
  bitClear(band,6);   // бит 6=0 для обычного времени кварца
                      // arg[2]-[7] are not used in this mode
  IRQflag=0;
  // послать
  Serial.println("...sending ATDD_POWER_UP with band set");
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(ATDD_POWER_UP);  
  // мы используем внешний кварц 32.768 кГц (установка бита 7)
  // без дополнительного времени ожидания (бит 6 в 0)
  Wire.write(band);
  Wire.endTransmission();  
  // ждем прерывания по включению
  while(!IRQflag);
  // ждать 2,5 мс, т.к. у нас прерывание по нарастающему фронту
  delayMicroseconds(2500);
  Serial.println("Si4844 ATDD_POWER_UP completed");
  Serial.println("...setting I2C speed to 50kHz");
  I2C_50kHz();
  delayMicroseconds(2000);
  // послать получение статуса и ждать 2 мс перед проверкой байта отклика
  Serial.println("...sending ATDD_GET_STATUS");
  get_status();
  Serial.println("...completed - tuner information is ready");
  /*
  // раскомментируйте, если хотите посмотреть байты отклика
  for(int x=0;x<4;x++)
  {
    Serial.print("Response byte ");
    Serial.print(x);
    Serial.print("= ");
    Serial.println(status_rsp[x],BIN);
  }
  */
  Serial.print("Band Mode is ");
  bandmode=status_rsp[1]>>6;
  Serial.print(bandmode);
  Serial.println(" [0=FM 1=AM 2=SW]");
  Serial.print("Band Index is ");
  band=status_rsp[1]&B00011111;
  Serial.print(band);
  Serial.println(" [0-19 FM / 20-24 AM / 25-40 SW]");
  // напечатать цифры частоты
  Serial.print("Tuner Frequency (MHz)= ");
  Serial.print(status_rsp[2]>>4);
  Serial.print(status_rsp[2]%16);
  Serial.print(status_rsp[3]>>4);
  Serial.print(".");
  Serial.println(status_rsp[3]%16);
  Serial.println();
  Serial.println("Chip information...");
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(GET_REV);  
  Wire.endTransmission();  
  delay(5);  // необходима реальная проверка CTS, но и это работает
  Wire.requestFrom(Si_ADDRESS, 0x09);
  for(int x=0;x<9;x++)
  {
    status_rsp[x]=Wire.read();
  }
  IRQflag=0;  
  Serial.print("Final two digits of part number: ");
  Serial.println(status_rsp[1]);
  Serial.print("Firmware revision: ");
  Serial.print((char)status_rsp[2]);
  Serial.print(".");
  Serial.println((char)status_rsp[3]);
  Serial.print("Component revision: ");
  Serial.print((char)status_rsp[6]);
  Serial.print(".");
  Serial.println((char)status_rsp[7]);
  Serial.print("Chip revision: ");
  Serial.println((char)status_rsp[8]);
  Serial.println();
  
  Serial.println("Entering tuning loop...");
  IRQflag=0;
  tuneloop:
  if(IRQflag)
  {
    Wire.beginTransmission(Si_ADDRESS);
    Wire.write(ATDD_GET_STATUS);  
    Wire.endTransmission();  
    Wire.requestFrom(Si_ADDRESS, 0x04);
    status_rsp[0]=Wire.read();
    status_rsp[1]=Wire.read();
    status_rsp[2]=Wire.read();
    status_rsp[3]=Wire.read();
    // напечатать цифры частоты
    Serial.print(status_rsp[2]>>4);
    Serial.print(status_rsp[2]%16);
    Serial.print(status_rsp[3]>>4);
    Serial.print(".");
    Serial.println(status_rsp[3]%16);
    IRQflag=0;
  }
  goto tuneloop;
}

void SiReset()
{
  // сбросить Si4844
  digitalWrite(RESETpin,LOW);
  delayMicroseconds(200);
  digitalWrite(RESETpin,HIGH);
  delayMicroseconds(200);
}

void get_status()
{
  // выполнять SiGetStatus, пока не получим корректную частоту в status_rsp
  // примечание: в тестировании мы не использовали цикл, но это хорошая практика
  // также, сюда может быть добавлена процедура выхода после определенного количества попыток
  Waitforgo:               
    SiGetStatus(); 
    if( ((bitRead(status_rsp[0],4)==0) || ( (status_rsp[2]==0) && (status_rsp[3]==0))) )
    {
      goto Waitforgo;       
    }
}

void SiGetStatus()
{
  // послать команду получения статуса и ждать 2 мс перед проверкой байта отклика
  Wire.beginTransmission(Si_ADDRESS);
  Wire.write(ATDD_GET_STATUS);  
  Wire.endTransmission();  
  delayMicroseconds(2000);
  // проверить байт отклика
  Wire.requestFrom(Si_ADDRESS, 0x04);
  for(int x=0;x<4;x++)
  {
    status_rsp[x]=Wire.read();
  }
}


// Примечание: Это функции установки скорости I2C для 8МГц , 3.3 В Pro Mini
// Для 16 МГц UNO или других плат Arduino их необходимо изменить

void I2C_10kHz()
{
  // установить скорость I2C на 10 кГц на чипе 8 МГц
  TWBR=98;
  TWSR|=bit(TWPS0);
}

void I2C_50kHz()
{
  // установить скорость I2C на 50 кГц на чипе 8 МГц
  TWBR=18;
  TWSR|=bit(TWPS0);
} 

// Функция обработки прерывания
void SiISR()
{
  IRQflag=1;
}



